﻿
###########################################################
#
# Implementace jednosměrně zřetězeného seznamu se zarážkou
#
# Tato verze nepoužívá iterátor, slouží především 
# k pochopení struktury a základních operací s jednosměrně
# zřetězeným seznamem.
###########################################################

import sys

###########################################################
# Třída prvku seznamu obsahuje pouze konstruktor 
# a metodu pro převod na znakový řetězec
class Prvek:
    """Prvek jednosměrně zřetězeného seznamu"""
    def __init__(self, data = None, další = None) -> None:
        """Konstruktor vytvoří prvek obsahující předanou hodnotu a předaný odkaz na další prvek"""
        self._data = data
        self._další = další

    def __str__(self) -> str:
        """Převede prvek na znakový řetězec"""
        return self._data.__str__()

###########################################################
#Třída seznamu
class Seznam:
    """Jednosměrně zřetězený seznam se zarážkou"""
    def __init__(self, *data) -> None:
        """Konstruktor vytvoří prázdný seznam
            a zaplní ho zadanými hodnotami"""
        self._hlava: Prvek = Prvek()
        self._zarážka: Prvek = self._hlava
        for údaj in data:
            self.vlož_na_konec(údaj)

    def __str__(self) -> str:
        """Převede seznam na řetězec"""
        pom = self._hlava
        výsledek = "("
        while pom != self._zarážka:
            výsledek += pom.__str__()
            pom = pom._další
            if pom != self._zarážka: výsledek += ", "
        return výsledek + ")"


# Vkládání nových hodnot do sezamu
    def vlož_na_konec(self, co):
        """Vloží novou hodnotu na konec seznamu"""
        self._zarážka._data = co
        self._zarážka._další = Prvek()
        self._zarážka = self._zarážka._další

    def vlož_za_daný_prvek(self, co, za_který: Prvek):
        """Vloží novou hodnotu za prvek určený iterátorem"""
        za_který._další = Prvek(co, za_který._další)


#Vyhledávání v seznamu
    def najdi(self, co)->Prvek:
        """Vyhledá prvek s danou hodnotou"""
        pom = self._hlava
        self._zarážka._data = co
        while pom._data != co:
            pom = pom._další
        return pom if pom != self._zarážka else None

# Odstraňování ze seznamu
    def odstraň_první(self):
        """Odstraní první prvek ze seznamu"""
        if self.prázdný(): 
            raise Exception("Odstraňování z prázdného seznamu")
        self._hlava = self._hlava._další
 
    def odstraň_za(self, za_kterým: Prvek):
        """Odstraní prvek za prvkem určeným iterátorem"""
        #pom = za_kterým._aktuální_prvek
        if za_kterým._další == self._zarážka:
            raise Exception("Odstraňování zarážky")
        za_kterým._další = za_kterým._další._další

    def odstraň(self, který: Prvek):
        """Odstraní prvek určený iterátorem"""
        if který._další == self._zarážka:
            self._zarážka = který
            který._další = None
        else:
            který._data = který._další._data
            self.odstraň_za(který)

    def vyprázdni(self):
        """Vyprázdní seznam"""
        self._hlava = self._zarážka

# Stav seznamu
    def prázdný(self) -> bool:
        """Vrátí True, je-li seznam prázdný"""
        return self._hlava == self._zarážka

# Indexování
    def _najdi_prvek_podle_indexu(self, index: int)->Prvek:
        pom: Prvek = self._hlava
        for i in range(0, index):
            if pom == self._zarážka: raise Exception("Index mimo rozsah")
            pom = pom._další
        return pom

    def __getitem__(self, index: int):
        pom: Prvek = self._najdi_prvek_podle_indexu(index)
        return pom._data

    def __setitem__(self, index, hodnota):
        pom: Prvek = self._najdi_prvek_podle_indexu(index)
        pom._data = hodnota

def main()->int: 
    s = Seznam(0, 10, 20, 30, 40, 50)
    print(s[5])
    s[5] = 7
    print(s)
    #s.vlož_na_konec(1)
    #s.vlož_na_konec(2)
    #print(s[0])
    #print(f"Výpis seznamu: {s}")
    #print(f"Hledána hodnota 2. Nalezeno: {s.najdi(2)}")
    #print(f"Hledána hodnota 3. Nalezeno: {s.najdi(3)}")
    #pom = s.najdi(1)
    #s.vlož_za_daný_prvek(3, pom)
    #print(f"Výpis seznamu: {s}")
    #s.odstraň_první()
    #print(f"Výpis seznamu: {s}")
    #pom = s.najdi(3)
    #s.odstraň_za(pom)
    #s.odstraň(pom)
    print(f"Výpis seznamu: {s}")
    return 0

if __name__ == "__main__":
    sys.exit(main())