﻿###########################################################
#
# Implementace jednosměrně zřetězeného seznamu se zarážkou
# a implementace iterátoru pro tento seznam
#
###########################################################

import sys

###########################################################
# Třída prvku seznamu obsahuje pouze konstruktor 
# a metodu pro převod na znakový řetězec
class Prvek:
    """Prvek jednosměrně zřetězeného seznamu"""
    def __init__(self, data = None, další = None) -> None:
        """Konstruktor vytvoří prvek obsahující předanou 
           hodnotu a předaný odkaz na další prvek"""
        self._data = data
        self._další = další

    def __str__(self) -> str:
        """Převede prvek na znakový řetězec"""
        return self._data.__str__()

###########################################################
class Iterátor:
    """Iterátor seznamu"""
    def __init__(self, aktuální_seznam, aktuální_prvek) -> None:
        """Konstruktor vytvoří iterátor ukazující na zadaný prvek zadaného seznamu"""
        self._aktuální_prvek = aktuální_prvek
        self._aktuální_seznam = aktuální_seznam

    # operátor == (iterátory jsou si rovny, když ukazují
    # na stejný prvek stejného seznamu)
    def __eq__(self, __value: object) -> bool:
        """Porovná dva iterátory"""
        return (self._aktuální_prvek == __value._aktuálníPrvek) and (self._aktuální_seznam == __value._aktuálníSeznam)

    def hodnota(self):
        """Vrátí hodnotu v prvku, na který iterátor ukazuje"""
        return self._aktuální_prvek._data

    def ulož_hodnotu(self, co):
        """Uloží do aktuálního prvku novou hodnotu"""
        self._aktuální_prvek._data = co

    def na_další(self):
        """Přesune iterátor na další prvek seznamu, 
        pokud existuje, jinak vyvolá výjimku"""
        if self._aktuální_prvek == self._aktuální_seznam._zarážka:
            raise Exception("Za konec seznamu")
        self._aktuální_prvek = self._aktuální_prvek._další

###########################################################
#Třída seznamu
class Seznam:
    """Jednosměrně zřetězený seznam se zarážkou"""
    def __init__(self, *data) -> None:
        """Konstruktor vytvoří prázdný seznam
            a zaplní ho zadanými hodnotami"""
        self._hlava: Prvek = Prvek()
        self._zarážka: Prvek = self._hlava
        self._velikost: int = 0
        for údaj in data:
            self.vlož_na_konec(údaj)


    def __str__(self) -> str:
        """Převede seznam na řetězec"""
        pom = self._hlava
        výsledek = "("
        while pom != self._zarážka:
            výsledek += pom.__str__()
            pom = pom._další
            if pom != self._zarážka: výsledek += ", "
        return výsledek + ")"

    # Vkládání nových hodnot do sezamu

    def vlož_na_začátek(self, co):
        """Vloží novou hodnotu na začátek seznamu"""
        self._hlava = Prvek(co, self._hlava)
        self._velikost += 1
           
    def vlož_na_konec(self, co):
        """Vloží novou hodnotu na konec seznamu"""
        self._zarážka._data = co
        self._zarážka._další = Prvek()
        self._zarážka = self._zarážka._další
        self._velikost += 1

    def vlož_za_daný_prvek(self, co, iter: Iterátor):
        """Vloží novou hodnotu za prvek určený iterátorem"""
        pom = iter._aktuální_prvek
        pom._další = Prvek(co, pom._další)
        self._velikost += 1

    #Vyhledávání v seznamu

    def najdi(self, co)->Iterátor:
        """Vyhledá prvek s danou hodnotou"""
        pom = self._hlava
        self._zarážka._data = co
        while pom._data != co:
            pom = pom._další
        return Iterátor(self, pom)

    # Odstraňování ze seznamu

    def odstraň_první(self):
        """Odstraní první prvek ze seznamu"""
        if self.prázdný(): 
            raise Exception("Odstraňování z prázdného seznamu")
        self._hlava = self._hlava._další
        self._velikost -= 1

    def odstraň_za(self, iter: Iterátor):
        """Odstraní prvek za prvkem určeným iterátorem"""
        pom = iter._aktuální_prvek
        if pom._další == self._zarážka:
            raise Exception("Odstraňování zarážky")
        pom._další = pom._další._další
        self._velikost -= 1

    def odstraň_určený(self, iter: Iterátor):
        """Odstraní prvek určený iterátorem"""
        pom = iter._aktuální_prvek
        if pom._další == self._zarážka:
            self._zarážka = pom
            pom._další = None
        else:
            pom._data = pom._další._data
            self.odstraň_za(iter)
        self._velikost -= 1

    def vyprázdni(self):
        """Vyprázdní seznam"""
        self._hlava = self._zarážka
        self._velikost = 0

    # Stav seznamu
    def prázdný(self) -> bool:
        """Vrátí True, je-li senam prázdný"""
        return self._hlava == self._zarážka

    def velikost(self) -> int:
        """Vrátí počet prvků seznamu"""
        return self._velikost
    
    # Metody vracející iterátory ukazující 
    # na začátek a za konec seznamu
    def začátek(self)->Iterátor:
        """Vrátí iterátor ukazující na hlavu seznamu"""
        return Iterátor(self, self._hlava)

    def konec(self)->Iterátor:
        """Vrátí iterátor ukazující za poslední prvek seznamu (na zarážku)"""
        return Iterátor(self, self._zarážka)

def main()->int:
    s = Seznam(5,6,4)
    s.vlož_na_začátek(5)
    s.vlož_na_začátek(7)
    s.vlož_na_konec(11)
    print(s)
    s.odstraň_první()
    print(s)
    it = s.najdi(5)
    it.ulož_hodnotu(45)
    it.na_další()
    it.ulož_hodnotu(88)
    print(s)
    s.vlož_za_daný_prvek(33, it)
    print(s)
    s.odstraň_za(it)
    s.odstraň_určený(it)
    print(s)
    return 0

if __name__ == "__main__":
    sys.exit(main())
