﻿###########################################################
#
#                     Hanojské věže
#
# Program ukazuje postup řešení úlohy "Hanojské věže".
# Počet disků je dán proměnnou počet_disků, nejvýše
# však hodnotou VÝŠKA_VĚŽE;
# doba mezi zobrazeními jednotlivých kroků je dána 
# proměnnou INTERVAL
#
# Nutno spustit z příkazové řádky příkazem
# 
# python HanojskéVěže.py
#
# neboť v některých IDE
# se výsledky nezobrazují správně (nemaže se okno)
###########################################################

import os
import sys
from time import sleep

global VÝŠKA_VĚŽE, INTERVAL, počet_disků, příznak, věže

VÝŠKA_VĚŽE = 10
INTERVAL = 0.5   #Interval čekání mezi zobrazeními stavu v sekundách
počet_disků = 4

####################################################
# Pomocné nástroje
# Fungují při spuštění z příkazové řádky,
# nefungují však např. ve vývojovém prostředí MS Visual Studio
####################################################

def smaž_okno():
    """Smaže konzolové okno."""
    os.system("CLS")

def čekej():
    """Program po zavolání této funkce počká na stisk libovolné klávesy."""
    print("")
    os.system("PAUSE")

#######################################################

def chyba(text):
    """Výpíše zprávu o chybě a ukončí program (funkce pro účely ladění)."""
    print("Chyba: ", end="")
    print(text)
    sys.exit(1)

#######################################################

class Věž:
    def __init__(self, kolik: int, jméno: str) -> None:
        """Konstruktor vytvoří věž s daným počtem disků a daným jménem"""
        self.stavba: list = [0]*(VÝŠKA_VĚŽE + 1)
        self.nastav_velikost(kolik)
        self.jméno: str = jméno

    def nastav_velikost(self, velikost):
        """Vyplní věž čísly znamenajícími velikost disku"""
        self.kolik: int = velikost
        for i in range(1, velikost + 1): 
            self.stavba[i] = velikost - i + 1

    def přenes(self, kam):
        """Přenese věž na zadané místo"""
        #Kontrola chyb při ladění
        if kam.kolik == VÝŠKA_VĚŽE: chyba("Přenos disku: přeplněný cíl")
        if self.kolik == 0: chyba("Přenos disku: prázdný zdroj")
        if (kam.kolik > 0) and (self.stavba[self.kolik] > kam.stavba[kam.kolik]): chyba("Přenos disku: chyba ve věži")
        # Vlastní přenos
        kam.kolik += 1
        kam.stavba[kam.kolik] = self.stavba[self.kolik]
        self.stavba[self.kolik] = 0
        self.kolik -= 1

    def nakresli(self):
        """Zobrazení jedné věže"""
        print(self.jméno, end=": ")
        for i in range(1, self.kolik + 1):
            print(self.stavba[i], end="")
        print("\n")
 
#######################################################

věže = [Věž(počet_disků, "výchozí věž  "), Věž(0, "cílová věž   "), Věž(0, "pomocná věž  ")]

def přemísti(KOLIK, A, B, C):
    """Přemístí KOLIK disků z A na B, C je pomocná věž"""
    if KOLIK == 1:
        A.přenes(B)
        zobraz()
    else:
        přemísti(KOLIK-1, A, C, B)
        přemísti(1, A, B, C)
        přemísti(KOLIK-1, C, B, A)

def zobraz():
    """Nakreslí všechny tři věže a chvilku počká."""
    global věže
    smaž_okno()
    for věž in věže: věž.nakresli()
    sleep(INTERVAL)

def main()->int:
    smaž_okno()
    zobraz()
    čekej()
    přemísti(počet_disků, věže[0], věže[1], věže[2])
    čekej()
    return 0
    
if __name__ == '__main__':
    sys.exit(main()) 





    


