﻿###########################################################
#
#                     Kalkulačka 
#  umí přiřazení, *, -, *, / a unární operátory + a -
###########################################################
#  syntax:
#
#  program:
#      KONEC
#      seznam_výrazů KONEC	//KONEC je konec vstupu, zde nemá význam
#
#  seznam_výrazů:
#      výraz TISK			//TISK je ';'
#      výraz TISK seznam_výrazů

#  výraz: 
#     člen - vyraz
#     člen + výraz
#     člen

#  člen:
#     primární_výraz / člen
#     primární_výraz * člen
#     primární_výraz

# primární_výraz: 
#    číslo 
#    jméno 
#    jméno = výraz
#    - primární_výraz
#    (primarni_vyraz)
#
# Slovem "Konec" ukončíme běh programu (na velikosti 
# písmen nezáleží).
###########################################################
# Nápad jsem převzal se Stroustrupovy knihy 
# The C++ Programming Language, 2. vydání
###########################################################
from enum import Enum
import math, sys

class TypSymbolu(Enum):
    """Výčtový typ pro reprezentaci typu symbolů (token) ve výrazech"""
    JMÉNO = 0
    ČÍSLO = 1
    KONEC = 2
    PLUS = ord('+')
    MINUS = ord('-')
    NÁSOBENÍ = ord('*')
    DĚLENÍ = ord('/')
    TISK = 7
    PŘIŘAZENÍ = ord('=')
    LEVÁ_ZÁVORKA = ord('(')
    PRAVÁ_ZÁVORKA = ord(')')
    TEČKA = ord('.')
    @classmethod
    def ze_znaku(cls, c: str):
        return cls(ord(c))

class Symbol:
    """Reprezentace symbolu (token)"""
    def __init__(self, _typ: TypSymbolu = TypSymbolu.TISK, _číselná_hodnota: float = 0, _řetězec: str = "") -> None:
        self.typ: TypSymbolu = _typ
        self.číselná_hodnota: float = _číselná_hodnota
        self.řetězec = _řetězec
    def __str__(self) -> str:
        """Převod symbolu na řetězec (pro ladění)"""
        return f"[{self.typ}, číselná hodnota: {self.číselná_hodnota}, řetězec: {self.řetězec}]"

global aktuální_symbol, počet_chyb, proměnná, pozice
aktuální_symbol: Symbol = Symbol()            # Poslední přečtený symbol
počet_chyb: int = 0
proměnná: dict = {"e": math.e, "pi": math.pi} # Proměnné (uložené jako dvojice jméno, hodnota)
pozice: int = 0                               # Aktuální pozice v řetězci "vstup" s výrazem

def chyba(text: str)->int:
    """Výpis zprávy o chybě"""
    global počet_chyb
    počet_chyb += 1
    print(text)
    return 1

def čti_symbol(vstup: str)->Symbol:
    """Přečte další symbol ze vstupního řetězce"""
    global pozice
    if pozice == len(vstup): return Symbol(TypSymbolu.KONEC)
    while pozice < len(vstup) and vstup[pozice] == ' ':
        if pozice == len(vstup): return Symbol(TypSymbolu.KONEC)
        pozice += 1
    if pozice == len(vstup): return Symbol(TypSymbolu.KONEC)
    match vstup[pozice]:
        case ';':
            pozice += 1
            return Symbol()
        case '+'|'-'|'*'|'/'|'='|'('|')':
            výsledek = Symbol(TypSymbolu.ze_znaku(vstup[pozice]), None, vstup[pozice])
            pozice += 1
            return výsledek
        case '0'|'1'|'2'|'3'|'4'|'5'|'6'|'7'|'8'|'9':
            text: str = vstup[pozice];
            while (pozice := pozice + 1) < len(vstup) and (vstup[pozice].isdigit() or vstup[pozice] == "."):
                text += vstup[pozice]
            return Symbol(TypSymbolu.ČÍSLO, float(text), text)
        case _:                              #jméno, chyba, konec
            if vstup[pozice].isalnum():
                text = vstup[pozice]
                while (pozice := pozice + 1) < len(vstup) and vstup[pozice].isalnum():
                    text += vstup[pozice]
                if text.upper() == "KONEC": return Symbol(TypSymbolu.KONEC)
                return Symbol(TypSymbolu.JMÉNO, None, text)
            else:
                chyba("Chybná položka: " + vstup[pozice])
                pozice +=1
                return Symbol()

def výraz(čti: bool, vstup: str)->float:
    """Vyhodnotí zadaný výraz"""
    global aktuální_symbol
    levý: float = člen(čti, vstup)
    while True:
        match aktuální_symbol.typ:
            case TypSymbolu.PLUS:
                levý += člen(True, vstup)
            case TypSymbolu.MINUS:
                levý -= člen(True, vstup)
            case _:
                break
    return levý

def člen(čti: bool, vstup: str)->float:
    """Vyhodnotí člen (term) v aktuálním výrazu"""
    global aktuální_symbol
    levý: float = primární(čti, vstup)
    while True:
        match aktuální_symbol.typ:
            case TypSymbolu.NÁSOBENÍ:
                levý *= primární(True, vstup)
            case TypSymbolu.DĚLENÍ:
                pom = primární(True, vstup)
                if pom != 0.0:
                    levý /= pom
                else:
                    chyba("Dělení nulou")
                    levý = 1
            case _:
                return levý

def primární(čti: bool, vstup: str)->float:
    """Vyhodnotí primární výraz v aktuálním výrazu"""
    global aktuální_symbol
    if čti:
        aktuální_symbol = čti_symbol(vstup)
    match aktuální_symbol.typ:
        case TypSymbolu.ČÍSLO:
            v = aktuální_symbol.číselná_hodnota
            aktuální_symbol = čti_symbol(vstup)
            return v
        case TypSymbolu.JMÉNO:
            jméno: str = aktuální_symbol.řetězec
            if jméno not in proměnná:
                proměnná[jméno] = 0
            aktuální_symbol = čti_symbol(vstup)
            if aktuální_symbol.typ == TypSymbolu.PŘIŘAZENÍ:
                v = výraz(True, vstup)
                proměnná[jméno] = v
            else: v = proměnná[jméno]
            return v
        case TypSymbolu.MINUS:
            return -primární(True, vstup)
        case TypSymbolu.LEVÁ_ZÁVORKA:
            v = výraz(True, vstup)
            if aktuální_symbol.typ != TypSymbolu.PRAVÁ_ZÁVORKA:
                chyba("Očekávána pravá závorka")
                return 1
            aktuální_symbol = čti_symbol(vstup)
            return v
        case _:
            chyba("Očekáván primární výraz")
            return 1
  
def main()->int:
    global aktuální_symbol, proměnná, pozice
    vstup = ""
    while True:
        try:
            vstup = input("Zadej výraz: ")
            pozice = 0
            aktuální_symbol = čti_symbol(vstup)
            if aktuální_symbol.typ == TypSymbolu.KONEC: break
            if aktuální_symbol.typ == TypSymbolu.TISK: continue
            print(výraz(False, vstup))
        except Exception as výjimka:
            print(výjimka)
            
if __name__ == "__main__":
    sys.exit(main())





