﻿###########################################################
#
# Implementace binárního vyhledávacího stromu
# Na metodu pro výpis je použit algoritmus pro odstranění 
# rekurze popsaný v odd. 4.2.1 skripta
###########################################################

from enum import Enum
import sys

###########################################################
# Prvek stromu
class Prvek:
    """Prvek binárního vyhledávacího stromu"""
    def __init__(self, data = None, levý = None, pravý = None) -> None:
        """Konstruktor vytvoří ůrvek obsahující předané hodnoty"""
        self._data = data
        self._levý = levý
        self._pravý = pravý
    
    def __str__(self) -> str:
        """Převod prvku na znakový řetězec"""
        return self._data.__str__()

###########################################################
# Výčtový typ pro určení míst v programu 
# použijeme ho místo návěští
class Návěští(Enum):
    L0 = 0
    L1 = 1
    KONEC = 3

###########################################################
# Třída strom
class Strom:
    """Třída binárního vyhledávacího stromu"""
    #######################################################
    # Metody tvořící rozhraní třídy Strom
    #######################################################
    def __init__(self) -> None:
        """Konstruktor vytvoří prázdný strom"""
        self._kořen: Prvek = None
        self._otec: Prvek = None
    
    def vlož(self, co):
        """ Vložení jedné hodnoty do stromu """
        if self.prázdný():
            self._kořen = Prvek(co)
        else:
            self._vlož_do_podstromu(co, self._kořen)

    def vlož_hodnoty(self, *hodnoty):
        """ Vložení skupiny hodnot do stromu"""
        for hodnota in hodnoty:
            self.vlož(hodnota)

    def prázdný(self) -> bool:
        """ Vrátí True, když strom neobsahuje žádné prvky, jinak vrátí False """
        return self._kořen is None

    def výpis(self):
        """ Výpis stromu na konzolu (především pro ladění programu)"""
        if not self._kořen: return
        else: self._výpis_podstromu_nerekurzivní(self._kořen)
        #else: self._výpis_podstromu(self._kořen)
        print('\n------')

    def najdi(self, co) -> Prvek:
        """ Vyhledání prvku se zadanou hodnotou ve stromě"""
        self._otec = None
        if self.prázdný(): return None
        aktuální = self._kořen
        while aktuální._data != co:
            self._otec = aktuální
            if aktuální._data < co:               
                aktuální = aktuální._pravý
            else:
                aktuální = aktuální._levý
            if aktuální is None: break
        return aktuální

    def odstraň(self, co):
        """ Odstraní zadanou hodnotu ze stromu"""
        odstraňovaný = self.najdi(co)
        if not odstraňovaný: 
            return
        else:
            self._odstraň(odstraňovaný)

    def vyprázdni(self):
        """Vyprázdní strom"""
        self._kořen = None

    #######################################################
    # Pomocné metody 
    #######################################################
    def _výpis_podstromu_nerekurzivní(self, kořenPodstromu: Prvek):
        """Pomocná metoda, jež vypíše podstrom určený kořenem - verze s odstraněnou rekurzí"""
        kam = Návěští.L0
        zásobník = []
        while True:
            match kam:
                case Návěští.L0:
                    if kořenPodstromu._levý: 
                        zásobník.append(kořenPodstromu)
                        kořenPodstromu = kořenPodstromu._levý                        
                    else:
                        kam = Návěští.L1
                    continue    
                case Návěští.L1:
                    print(kořenPodstromu._data, end=" ")
                    if kořenPodstromu._pravý: 
                        kořenPodstromu = kořenPodstromu._pravý
                        kam = Návěští.L0
                    else:
                        kam = Návěští.KONEC
                    continue
                case Návěští.KONEC:
                    if zásobník:
                        kam = Návěští.L1
                        kořenPodstromu = zásobník.pop()
                        continue
                    else:
                        break

    def _výpis_podstromu(self, kořenPodstromu: Prvek):
        """Pomocná metoda, jež vypíše podstrom určený kořenem - původní rekurzivní verze"""
        if kořenPodstromu._levý:
            self._výpis_podstromu(kořenPodstromu._levý)
        print(kořenPodstromu._data, end=" ")
        if kořenPodstromu._pravý:
            self._výpis_podstromu(kořenPodstromu._pravý)

    def _vlož_do_podstromu(self, co, kořen_podstromu: Prvek):
        """Pomocná metoda, jež vloží zadanou hodnotu do podstromu určeného kořenem"""
        if kořen_podstromu._data == co: return
        if co > kořen_podstromu._data:
            if not kořen_podstromu._pravý:
                kořen_podstromu._pravý = Prvek(co)
            else:
                self._vlož_do_podstromu(co, kořen_podstromu._pravý)
        else:
            if not kořen_podstromu._levý:
                kořen_podstromu._levý = Prvek(co)
            else:
                self._vlož_do_podstromu(co, kořen_podstromu._levý)

    def _odstraň(self, odstraňovaný: Prvek):
        """Pomocná metoda, jež odstraní zadaný prvek"""
        if (odstraňovaný._levý is None) and (odstraňovaný._pravý is None): 
            self._odstraň_list(odstraňovaný)
        elif (odstraňovaný._levý is None) or (odstraňovaný._pravý is None):
            self._odstraň_s_jedním_synem(odstraňovaný)
        else:
            self._odstraň_se_dvěma_syny(odstraňovaný)

    def _odstraň_list(self, odstraňovaný: Prvek):
        """Pomocná metoda, jež odstraní zadaný list stromu"""
        if odstraňovaný is self._kořen:
            self._kořen = None
        else:
            if odstraňovaný._data > self._otec._data:
                self._otec._pravý = None
            else:
                self._otec._levý = None

    def _odstraň_s_jedním_synem(self, odstraňovaný: Prvek):
        """Pomocná metoda, jež odstraní zadaný prvek s jedním synem"""
        syn = odstraňovaný._levý if odstraňovaný._pravý is None \
              else odstraňovaný._pravý
        if odstraňovaný is self._kořen:
            self._kořen = syn
        else:
           if self._otec._levý == odstraňovaný:
                self._otec._levý = syn
           else:
                self._otec._pravý = syn

    def _odstraň_se_dvěma_syny(self, odstraňovaný: Prvek):
        """Pomocná metoda, jež odstraní  zadaný prvek se dvěma syny"""
        zástupce = self._najdi_zástupce(odstraňovaný)
        odstraňovaný._data = zástupce._data
        self._odstraň(zástupce)

    def _najdi_zástupce(self, za_koho: Prvek) -> Prvek:
        """Pomocná metoda, jež najde zástupce za odstraňovaný prvek se dvěma syny"""
        self._otec = za_koho
        zástupce = za_koho._levý
        while zástupce._pravý:
            self._otec = zástupce
            zástupce = zástupce._pravý
        return zástupce


def main()->int:
    s = Strom()
    #print("prázdný" if s.prázdný() else "neprázdný")
    #s.výpis()
    #s.vlož(55)
    #s.vlož(44)
    #s.výpis()
    #s.odstraň(44)
    #s.výpis()
    #s.odstraň(55)
    #s.výpis()
    #s.vlož(4)
    #s.vlož(6)
    #s.vlož(11)
    s.vlož_hodnoty(5, 3, 4, 6, 2, 8, 9, 1, 11, 10, 7, -1)
    #s.vlož_hodnoty(5, 6, 8, 9, 11, 10, 7)
    s.výpis()
    #ten = s.najdi(11)
    #print(ten)
    #s.odstraň(50)
    #s.výpis()
    return 0

if __name__ == '__main__':
    sys.exit(main())
