﻿########################################################
# 
#                   Hoarův algoritmus
#
# Program najde v poli A prvek, který by byl na místě
# s indexem KOLIKÁTÝ, kdyby bylo seřazeno podle velikosti.
# Pole A obsahuje hodnoty 0 .. 19 v náhodném pořadí, 
# takže program by měl vypsat hodnotu rovnou KOLIKÁTÝ
# Vedle toho vypíše pole před hledáním, po dokončení 
# algoritmu a po seřazní
########################################################


from random import shuffle
import sys

global KOLIKÁTÝ
KOLIKÁTÝ: int = 11

def prohoď(A, i, j):
    A[i], A[j] = A[j], A[i]

def _rozděl(A: list, levý: int, pravý: int, index_pivotu: int)->int:
    """Rozdělí pole podle zadaného pivotu (jako u quicksortu)"""
    x = A[index_pivotu]
    prohoď(A, index_pivotu, levý)
    i, k = levý + 1, pravý + 1
    while i < k:
        if A[i] < x:
            i += 1
        else:
            k -= 1
            prohoď(A, k, i)
    prohoď(A, levý, k - 1)
    return k - 1

def najdi(A: list, k: int):
    """Hoarův algoritmus vyhledání k-tého prvku podle velikosti (indexováno od 0)"""
    levý, pravý = 0, len(A) - 1
    nalezeno = _rozděl(A, levý, pravý, k)
    while nalezeno != k:
        if nalezeno > k:
            pravý = nalezeno - 1
        else:
            levý = nalezeno + 1
        nalezeno = _rozděl(A, levý, pravý, k)
    return A[k]

def příprava(A: list):
    """Naplní pole hodnotami a náhodně je promíchá"""
    for i in range(0, len(A)):
        A[i] = i
    shuffle(A)

def main()->int:
    A:list = [0]*20
    příprava(A)
    print(f"Zadané pole:      {A}")
    print("Hledáme %d. prvek (indexováno od 0)" % KOLIKÁTÝ)
    print("Nalezeno: %d" % najdi(A, KOLIKÁTÝ))
    print(f"Pole po skončení: {A}")
    list.sort(A)      # Seřadíme pole pomocí vestavěné metody
    print(f"Seřazené pole:    {A}")
    return 0

if __name__ == '__main__':
    sys.exit(main())
