﻿###########################################################
# 
#            Introspektivní třídění (introsort)
#
#  Využívá quicksort, třídění haldou a třídění vkládáním
###########################################################

from math import log2
import time
import sys
global délka
délka: int = 5

# Přímé vkládání s posunutím všech prvků najednou
def přímé_vkládání(A: list):
	"""Třídění přímým vkládáním pro optimalizaci quicksortu"""
	for i in range(1, len(A)):
		j: int = i
		while (j > 0) and(A[j-1] > A[i]):
			j -= 1
		if j != i:
			pom = A[i]
			for k in range(i, j, -1):
				A[k] = A[k-1]
			A[j] = pom

###########################################################
#
# Třídění úseku pole haldou 
#
# Pomocná funkce
# "počátek" je začátek tříděného úseku, "pravý" je konec haldy
# Od podobné funkce v příkladu třídění haldou se liší jen tím, že
# indexy jsou posunuty, aby počáteční prvek tříděného úseku 
# měl index 0
def _vlož_do_haldy(A: list, levý: int, pravý: int, počátek: int):
	"""Pomocná funkce - vloží nový prvek do haldy"""
	i: int = levý - počátek
	j: int = 2*i + 1
	x = A[i + počátek]                 
	while j <= pravý - počátek:        
		if j < pravý - počátek:        
			if A[j + počátek] < A[j + 1 + počátek]: j += 1
		if x >= A[j + počátek]: break   
		A[i + počátek] = A[j + počátek]
		i = j
		j = 2*i + 1
	A[i + počátek] = x


# Vlastní třídění úseku 
# od příkladu třídění haldou se liší jen výpočet meze počáteční 
# haldy a volání pomocné funkce
def _třídění_úseku_haldou(A: list, počátek: int, konec: int):
	"""Setřídí zadaný úsek pole algoritmem třídění haldou"""
	print("halda")
	levý = ((konec - počátek) // 2) + 1 + počátek           
	pravý = konec                            
	while levý > počátek:                      
		levý -= 1
		_vlož_do_haldy(A, levý, pravý, počátek)		
	while pravý > počátek:            
		A[levý], A[pravý] = A[pravý], A[levý]
		pravý -= 1
		_vlož_do_haldy(A, levý, pravý, počátek)

#rozdělení podle pivotu
def _rozděl(A: list, levý: int, pravý: int, index_pivotu: int)->int:
	"""Rozdělí zadaný úsek pole podle zadaného pivotu"""
	x = A[index_pivotu]
	A[index_pivotu], A[levý] = A[levý], A[index_pivotu] #Prohoď pivot s počátečním prvkem
	i = levý + 1
	k = pravý + 1
	while i < k:                  #Je-li prvek menší než pivot,
		if A[i] < x:              #nech ho být,
			i += 1
		else:                     #jinak ho "odlož" na konec pole
			k -= 1
			A[k], A[i] = A[i], A[k]
	A[levý], A[k-1] = A[k-1], A[levý]
	return k - 1


def _intro(A: list, levý: int, pravý: int, hloubka: int):
	"""Vlastní implementace introspektivního třídění; 
	   parametr "hloubka slouží k hlídání hloubkuy rekurze"""
	while pravý - levý > délka:
		if hloubka == 0:          
			_třídění_úseku_haldou(A, levý, pravý - levý + 1)
			return 
		indexPivotu = _rozděl(A, levý, pravý, (levý + pravý) // 2)
		hloubka -= 1
		_intro(A, indexPivotu + 1, pravý, hloubka)
		pravý = indexPivotu - 1

def introsort(A: list):
	_intro(A, 0, len(A) - 1, int(2*log2(len(A))))
	přímé_vkládání(A)

def main()->int:
	a = [5,2,4,6,8,1,7,3,0,10,15,9,11,14,13,12,20,19,18,17,16]
	print("Zadané pole:")
	print(a)
	time.sleep(2)
	#os.system("cls")   # V protředí MS Visual Studio tento příkaz nefunguje
	introsort(a)
	print("Seřazené pole:")
	print(a)
	return 0

if __name__ == '__main__':
    sys.exit(main())