﻿###########################################################
#
#              Rychlé třídšní - quicksort
#               
# Program obsahuje dvě implementace s růzmými rozdělovacími 
# algoritmy. Druhá implementace také ukazuje optimalizaci -
# úseky kratší než zadaná délka se již dále nerozdělují
# a na pole s takovýmito úseky se použije třídění vkládáním
###########################################################
import sys
from this import s

global délka
délka: int = 10    # Úseky kratší než délka se třídí vkládáním

###########################################################
# První implementace
########################################################### 
def _třídění(A:list, levý: int, pravý: int):
	"""Vlastní implementce quicksortu"""
	i, j = levý, pravý
	pivot: int = A[(i + j) // 2]
	while True:
		while A[i] < pivot: i += 1  #Tyto dva cykly while hledají prvky,
		while A[j] > pivot: j -= 1  #které se navzájem vymění
		if i <= j:
			A[i], A[j] = A[j], A[i] #Výměna
			i += 1
			j -= 1
		if i > j: break
	if levý < j: _třídění(A, levý, j) #Prvek x na svém místě
	if pravý > i: _třídění(A, i, pravý) #Třídí se ostatní

def quicksort(A: list):
	"""Neoptimalizovaný quicksort s prvním rozdělovacím algoritmem"""
	_třídění(A, 0, len(A)-1)

###########################################################
# První implementace, nerekurzivní verze
########################################################### 
def quicksort_n(A):
	zásobník = [0, len(A)-1]
	while True:
		pravý, levý = zásobník.pop(), zásobník.pop()
		while True:
			i, j = levý, pravý
			pom = (levý+ pravý) // 2
			pivot = A[(levý + pravý) // 2]
			while True:
				while A[i] < pivot: i += 1
				while A[j] > pivot: j -= 1
				if i <= j: 
					A[i], A[j] = A[j], A[i]
					i += 1
					j -= 1
				if i > j: break
			if i < pravý:
				zásobník.append(i)
				zásobník.append(pravý)
			pravý = j
			if levý >= pravý: break
		if not zásobník: break

###########################################################
#Alternativní implementace s omezením délky intervalu zdola
###########################################################

# Přímé vkládání pro optimalizaci quicksortu
def přímé_vkládání(A: list):
	"""Přímé vkládání s posunutím všech prvků najednou"""
	for i in range(1, len(A)):
		j = i
		while (j > 0) and(A[j-1] > A[i]):
			j -= 1
		if j != i:
			pom = A[i]
			for k in range(i, j, -1):
				A[k] = A[k-1]
			A[j] = pom
  			
def _rozděl(A: list, levý: int, pravý: int, index_pivotu: int)->int:
	"""Alternativní rozdělovací algoritmus pro quicksort"""
	x = A[index_pivotu]
	A[index_pivotu], A[levý] = A[levý], A[index_pivotu] #Prohoď pivot s počátečním prvkem
	i = levý + 1
	k = pravý + 1
	while i < k:                  #Je-li prvek menší než pivot,
		if A[i] < x:              #nech ho být,
			i += 1
		else:                     #jinak ho "odlož" na konec pole
			k -= 1
			A[k], A[i] = A[i], A[k]
	A[levý], A[k-1] = A[k-1], A[levý]
	return k - 1

def _qsort(A: list, levý: int, pravý: int):
	"""Třídění rozdělováním, které je záladem quicksortu, 
	   dokud jsou intervaly delší než zadaná hodnota"""
	while pravý - levý > délka:
		index_pivotu:int = _rozděl(A, levý, pravý, (levý + pravý) // 2)
		_qsort(A, index_pivotu + 1, pravý)
		pravý = index_pivotu - 1

def quicksort_1(A: list):
	"""Optimalizovaný quicksort s druhým rozdělovacím algoritmem"""
	_qsort(A, 0, len(A) - 1)
	přímé_vkládání(A)

def main()->int:
	a = [5,2,4,6,8,1,7,3,0,10,15,9,11,14,13,12,20,19,18,17,16]
	#quicksort(a)
	quicksort_n(a)
	#quicksort_1(a)
	print(a)
	return 0

if __name__ == '__main__':
    sys.exit(main())