﻿###########################################################
#              Třídění polí slučováním 
#
# Podle https://www.geeksforgeeks.org/python-program-for-merge-sort/

from random import shuffle
import sys

###########################################################
#
# Pomocná funkce
# 
def _slučování(pole_A: list, levý: int, střed: int, pravý: int)->None:
	"""Sloučí dvě podpole, která jsou součástí 
       tříděného pole_A[]. První podlole je pole_A[levý..střed], druhé 
	   je pole_A[střed+1..pravý]"""
	n1: int = střed - levý + 1  # Délky podpolí
	n2: int = pravý - střed

	# Vytvoříme dočasná pomocná pole_B[] 
	# a pole_C[] a okopírujeme do nich data 
	pole_B = pole_A[levý : levý + n1]
	pole_C = pole_A[střed + 1 : střed + 1 + n2]

	# Sloučíme pomocná pole zpět do pole_A[l..r].
	# Počáteční index prvního a druhého podpole a sloučeného pole
	i, j, k = 0, 0, levý 
	while i < n1 and j < n2:
		if pole_B[i] <= pole_C[j]:
			pole_A[k] = pole_B[i]
			i += 1
		else:
			pole_A[k] = pole_C[j]
			j += 1
		k += 1

	# Okopírujeme zbylé prvky pole_B[], jsou-li jaké
	while i < n1:
		pole_A[k] = pole_B[i]
		i += 1
		k += 1

	# Okopírujeme zbylé prvky pole_C[], jsou-li jaké
	while j < n2:
		pole_A[k] = pole_C[j]
		j += 1
		k += 1


# Parametry levý, pravý představují levý, resp. prevý index tříděného podpole
def _třídění_slučováním_impl(pole: list, levý: int, pravý: int)->None:
	""" Třídění polí přímým slučováním - vlastní implementace"""
	if levý < pravý:
		střed: int = levý+(pravý-levý)//2   # Totéž co (levý+pravý)//2

		# Setřiď první a druhou polovinu
		_třídění_slučováním_impl(pole, levý, střed)
		_třídění_slučováním_impl(pole, střed+1, pravý)
		_slučování(pole, levý, střed, pravý)

def třídění_slučováním(pole: list)-> None:
	"""Třídění pole přímým slučováním"""
	_třídění_slučováním_impl(pole, 0, len(pole) - 1)


def vytvoř_pole(délka: int)->list:
	"""Vytvoří pole zadané délky s náhodně promíchanými prvky 0 .. délka-1""" 
	výsledek = [i for i in range(0, délka)]
	shuffle(výsledek)
	return výsledek
 
def main() -> int:
	pole: list = vytvoř_pole(31)
	print(f"Zadané pole  : {pole}")
	třídění_slučováním(pole)
	print(f"Seřazené pole: {pole}")
	return 0

if __name__ == "__main__":
	sys.exit(main())

