﻿###########################################################
#
#           Implementace vybraných třídicích metod
#
#  Přímé vkládání (3 různé implementace), binární vkládání,
#  třídění výběrem, bubliknové třídění, třídění přetřásáním
###########################################################

import sys

###########################################################
# Přímé vkládání s postupnými výměnami -- jednodušší kód, ale 
# pomalejší
#
def přímé_vkládání_0(A: list):
    """Přímé vkládání s postupnými výměnami"""
    for i in range(1, len(A)):
        j = i
        while (j > 0) and (A[j-1] > A[j]):
            pom = A[j]
            A[j] = A[j-1]
            A[j-1] = pom
            j -= 1


###########################################################
# Přímé vkládání s cyklickým posunutím všech prvků
#  
def přímé_vkládání(A: list):
    """Přímé vkládání s cyklickým posunutím všech prvků"""
    for i in range(1, len(A)):
        j = i
        while (j > 0) and(A[j-1] > A[i]):
            j -= 1
        if j != i:
            pom = A[i]
            for k in range(i, j, -1):
                A[k] = A[k-1]
            A[j] = pom
    
###########################################################
# Přímé vkládání přehledněji implementované (skládá se ze 3 funkcí)
#

# Vlastní třídění
def přímé_vkládání_1(A: list):
    """Přímé vkládání přehledněji implementované"""
    for i in range(1, len(A)):
        pom = A[i]
        pozice = _najdi_místo(A, i, pom)
        if pozice != i:
            _zařaď_na_místo(A, pom, i, pozice)

# Pomocné funkce: nalezení místa pro zařazovanou hodnotu 
# a její zařazení na toto místo
def _najdi_místo(A: list, n: int, pom: int) -> int:
    """Pomocná funkce pro nalezení správného místa prvku
       Vrátí index, kam prvek patří"""
    while (n > 0) and (A[n-1] > pom):
        n -= 1
    return n

def _zařaď_na_místo(A: list, co, odkud: int, kam: int ):
    """Zařazení prvku na správné místo ve tříděném poli"""
    for k in range(odkud, kam, -1):
        A[k] = A[k-1]
    A[kam] = co

###########################################################
#Binární vkládání
#
def binární_vkládání(A: list):
    for i in range(1, len(A)):
        pom = A[i]
        pozice = _najdi_místo_binárně(A, i, pom)
        if pozice != i:
            _zařaď_na_místo(A, pom, i, pozice)

# Pomocná funkce pro vyhledání místa pro zařazovanou hodnotu
def _najdi_místo_binárně(A: list, n: int, x) -> int:
    """Binární vyhledání správného místa pro binární vkládání"""
    levý, pravý = 0, n
    while levý < pravý:
        střed = (levý + pravý) // 2
        if x <= A[střed]: pravý = střed
        else: levý = střed + 1
    return levý

###########################################################
#Třídění výběrem
#
def výběr(A: list):
    for i in range(0, len(A)):
        nejmenší = najdi_nejmenší(A, i)
        if nejmenší != i: A[i], A[nejmenší] = A[nejmenší], A[i]

# pomocná funkce: Nalezení indexu nejmenšího prvku v daném úseku pole
def najdi_nejmenší(A: list, od: int)->int:
    """Nalezení indexu nejmenšího prvku v daném úseku pole"""
    nejmenší: int = od
    for i in range(nejmenší + 1, len(A)):
        if A[i] < A[nejmenší]: nejmenší = i
    return nejmenší

###########################################################
#Bublinkové třídění 
#
def bublinky_0(A: list):
    """Bublinové třídění, nejjednodušší verze - nesleduje počet a polohu výměn"""
    for i in range(1, len(A)):
        for j in range(len(A)-1, i-1, -1):
            if A[j] < A[j-1]: A[j], A[j-1] = A[j-1], A[j]


#Lepší verze - sleduje počet a polohu výměn
def bublinky(A: list):
    """Bublinové třídění, lepší verze - sleduje počet a polohu výměn"""
    poslední_výměna: int = 1
    while True:
        počet_výměn: int = 0
        pom = poslední_výměna
        for j in range(len(A)-1, pom-1, -1):
            if A[j] < A[j-1]: 
                A[j], A[j-1] = A[j-1], A[j]
                počet_výměn +=1
                poslední_výměna = j
        if počet_výměn == 0: break

###########################################################
#Třídění přetřásáním
#
def třesení(A: list):
    """Třídění přetřásáním"""
    poslední_výměna = len(A) - 1
    levý, pravý = 1, len(A) - 1
    while True:                               #Cyklus přes přibližující se konce procházeného intervalu
        for j in range(pravý, levý-1, -1):    #Průchod od konce k počátku
            if A[j-1] > A[j]:
                A[j], A[j-1] = A[j-1], A[j]
                poslední_výměna = j            #Index poslední výměny
        levý = poslední_výměna + 1             #Příště procházíme jen od l
        for j in range(levý, pravý + 1, 1):   #Průchod od počátku ke konci
            if A[j-1] > A[j]:
                A[j], A[j-1] = A[j-1], A[j]
                poslední_výměna = j            #Index poslední výměny
        pravý = poslední_výměna - 1            #Příště procházíme jen od pravého
        if levý > pravý: break                #Když se levý a pravý konec setkají, končíme


def main()->int:
    #tříděné pole: zpřeházená čísla 0 .. 20 
    a = [5,2,4,6,8,1,7,3,0,10,15,9,11,14,13,12,20,19,18,17,16]

    přímé_vkládání_0(a)
    #přímé_vkládání(a)
    #přímé_vkládání_1(a)
    #binární_vkládání(a)
    #výběr(a)
    #bublinky_0(a)
    #bublinky(a)
    #třesení(a)
    print(a)
    return 0

if __name__ == "__main__":
    sys.exit(main())