﻿###########################################################
#
# Třídění jednosměrně zřetězeného seznamu se zarážkou
#       přirozeným slučováním (mergesort)
# 
# Třída Seznam je stejná jako v příkladu ke 2. kapitole, 
# navíc je pouze metoda seřaď() a pomocné metody pro 
# rozdělování a slučování běhů
#
###########################################################

import sys
from random import shuffle    # Pro přípravu dat ke třídění

###########################################################
# Třída prvku seznamu obsahuje pouze konstruktor 
# a metodu pro převod na znakový řetězec
class Prvek:
    """Prvek jednosměrně zřetězeného seznamu"""
    def __init__(self, data = None, další = None) -> None:
        """Konstruktor vytvoří prvek obsahující předanou 
           hodnotu a předaný odkaz na další prvek"""
        self._data = data
        self._další = další

    def __str__(self) -> str:
        """Převede prvek na znakový řetězec"""
        return self._data.__str__()

###########################################################
class Iterátor:
    """Iterátor seznamu"""
    def __init__(self, aktuální_seznam, aktuální_prvek) -> None:
        """Konstruktor vytvoří iterátor ukazující na zadaný prvek zadaného seznamu"""
        self._aktuální_prvek = aktuální_prvek
        self._aktuální_seznam = aktuální_seznam

    # operátor == (iterátory jsou si rovny, když ukazují
    # na stejný prvek stejného seznamu)
    def __eq__(self, __value: object) -> bool:
        """Porovná dva iterátory"""
        return (self._aktuální_prvek == __value._aktuálníPrvek) and (self._aktuální_seznam == __value._aktuálníSeznam)

    def hodnota(self):
        """Vrátí hodnotu v prvku, na který iterátor ukazuje"""
        return self._aktuální_prvek._data

    def ulož_hodnotu(self, co):
        """Uloží do aktuálního prvku novou hodnotu"""
        self._aktuální_prvek._data = co

    def na_další(self):
        """Přesune iterátor na další prvek seznamu, 
        pokud existuje, jinak vyvolá výjimku"""
        if self._aktuální_prvek == self._aktuální_seznam._zarážka:
            raise Exception("Za konec seznamu")
        self._aktuální_prvek = self._aktuální_prvek._další

###########################################################
#Třída seznamu
class Seznam:
    """Jednosměrně zřetězený seznam se zarážkou"""
    def __init__(self, *data) -> None:
        """Konstruktor vytvoří prázdný seznam
            a zaplní ho zadanými hodnotami"""
        self._hlava: Prvek = Prvek()
        self._zarážka: Prvek = self._hlava
        self._velikost: int = 0
        for údaj in data:
            self.vlož_na_konec(údaj)


    def __str__(self) -> str:
        """Převede seznam na řetězec"""
        pom = self._hlava
        výsledek = "("
        while pom != self._zarážka:
            výsledek += pom.__str__()
            pom = pom._další
            if pom != self._zarážka: výsledek += ", "
        return výsledek + ")"

    # Vkládání nových hodnot do sezamu

    def vlož_na_začátek(self, co):
        """Vloží novou hodnotu na začátek seznamu"""
        self._hlava = Prvek(co, self._hlava)
        self._velikost += 1
           
    def vlož_na_konec(self, co):
        """Vloží novou hodnotu na konec seznamu"""
        self._zarážka._data = co
        self._zarážka._další = Prvek()
        self._zarážka = self._zarážka._další
        self._velikost += 1

    def vlož_za_daný_prvek(self, co, iter: Iterátor):
        """Vloží novou hodnotu za prvek určený iterátorem"""
        pom = iter._aktuální_prvek
        pom._další = Prvek(co, pom._další)
        self._velikost += 1

    #Vyhledávání v seznamu

    def najdi(self, co)->Iterátor:
        """Vyhledá prvek s danou hodnotou"""
        pom = self._hlava
        self._zarážka._data = co
        while pom._data != co:
            pom = pom._další
        return Iterátor(self, pom)

    # Odstraňování ze seznamu

    def odstraň_první(self):
        """Odstraní první prvek ze seznamu"""
        if self.prázdný(): 
            raise Exception("Odstraňování z prázdného seznamu")
        self._hlava = self._hlava._další
        self._velikost -= 1

    def odstraň_za(self, iter: Iterátor):
        """Odstraní prvek za prvkem určeným iterátorem"""
        pom = iter._aktuální_prvek
        if pom._další == self._zarážka:
            raise Exception("Odstraňování zarážky")
        pom._další = pom._další._další
        self._velikost -= 1

    def odstraň_určený(self, iter: Iterátor):
        """Odstraní prvek určený iterátorem"""
        pom = iter._aktuální_prvek
        if pom._další == self._zarážka:
            self._zarážka = pom
            pom._další = None
        else:
            pom._data = pom._další._data
            self.odstraň_za(iter)
        self._velikost -= 1

    def vyprázdni(self):
        """Vyprázdní seznam"""
        self._hlava = self._zarážka
        self._velikost = 0

    # Stav seznamu
    def prázdný(self) -> bool:
        """Vrátí True, je-li senam prázdný"""
        return self._hlava == self._zarážka

    def velikost(self) -> int:
        """Vrátí počet prvků seznamu"""
        return self._velikost
    
    # Metody vracející iterátory ukazující 
    # na začátek a za konec seznamu
    def začátek(self)->Iterátor:
        """Vrátí iterátor ukazující na hlavu seznamu"""
        return Iterátor(self, self._hlava)

    def konec(self)->Iterátor:
        """Vrátí iterátor ukazující za poslední prvek seznamu (na zarážku)"""
        return Iterátor(self, self._zarážka)

    # Třídění seznamu přirozeným slučováním
    
    def seřaď(self)->None:
        """Implementace přirozeného slučování (mergesort) pro spojový seznam"""
        B = Seznam()                 # Pomocné seznamy
        C = Seznam()
        počet_běhů: int = 0
        while True:
            self._rozděl(B, C)
            počet_běhů = self._sluč(B, C)
            if počet_běhů == 1: break

    def _rozděl(self, B, C)->None:
        """Rozdělí běhy z aktuálního seznamu do seznamů B a C"""
        while True:
            self._zkopíruj_běh(self, B)
            if not self.prázdný():
                self._zkopíruj_běh(self, C)
            if self.prázdný(): break
        return

    def _zkopíruj_běh(self, X, Y)->None:
        """Okopíruje jeden běh ze seznamu X do Y"""
        konec_běhu: bool = False              # V cyklu kopíruje prvky,
        while True:                           # dokud nedojde na konec běhu
            konec_běhu = self._zkopíruj_prvek(X, Y)
            if konec_běhu: break

    def _zkopíruj_prvek(self, X, Y)->bool:
        """Okopíruje jeden prvek ze seznamu X do Y"""
        pom = X._hlava
        X._hlava = X._hlava._další
        Y._zarážka._data = pom._data
        Y._zarážka._další = pom
        Y._zarážka = pom
        if X.prázdný(): return True
        else: return pom._data > X._hlava._data

    def _sluč(self, B, C)->int:
        """Sloučí seznamy B a C do aktuálního seznamu"""
        počet_běhů: int = 0
        while (not B.prázdný()) and (not C.prázdný()):
            self._sluč_běh(B, C)
            počet_běhů += 1
        while not B.prázdný():
            self._zkopíruj_běh(B, self)
            počet_běhů += 1
        while not C.prázdný():
            self._zkopíruj_běh(C, self)
            počet_běhů += 1
        return počet_běhů

    def _sluč_běh(self, B, C):
        """Sloučí aktuální běh z B s aktuálním během z C a výsledek 
           zapíše do aktuálního seznamu"""
        konec_běhu: bool = False
        while True:
            if B._hlava._data < C._hlava._data:
                konec_běhu = self._zkopíruj_prvek(B, self)
                if konec_běhu: self._zkopíruj_běh(C, self)
            else:
                konec_běhu = self._zkopíruj_prvek(C, self)
                if konec_běhu: self._zkopíruj_běh(B, self)
            if konec_běhu: break

def příprava(seznam: Seznam, n: int):
    """Naplní seznam náhodně promíchanými hodnotami 0 .. n"""
    A = list()
    for i in range(0, n+1):
        A.append(i)
    shuffle(A)
    for i in range (0, n+1):
        seznam.vlož_na_konec(A[i])


def main()->int:
    s = Seznam()
    příprava(s, 30)
    print(s)
    s.seřaď()
    print(s)
    return 0

if __name__ == "__main__":
    sys.exit(main())

