﻿###########################################################
#
#              Topologické třídění
#
# Program očekává v souboru data.txt dvojice hodnot x a y 
# takové, že x < y. (Pozor, jde o částečné uspořádání.)
# Vypíše jedno z možných seřazení vstupních hodnot v pořadí, 
# v němž pro každý prvek z platí, že před ním jsou prvky menší
# nebo prvky, pro které není relace < s prvkem z definována
###########################################################

import sys

# Hlavní a pomocný prvek seznamu
class Hlavní:
    """Hlavní prvek seznamu"""
    def __init__(self, _klíč = None, _další = None, _pomocný = None) -> None:
        """Konstruktor vytvoří hlavní prvek seznamu se zadanými daty"""
        self.klíč = _klíč
        self.počet_předchůdců = 0
        self.další = _další
        self.pomocný = _pomocný

    def __str__(self) -> str:
        """Vytvoří řetězec obsahující data aktuálního prvku"""
        výsledek: str = str(self.klíč) + ": Počet předchůdců: " + str(self.počet_předchůdců) + "; následovníci: "
        pom = self.pomocný
        while pom != None:
            výsledek += str(pom.následovník.klíč) + ", "
            pom = pom.další
        return výsledek 
        

class Pomocný:
    """Prvek pomocného seznamu"""
    def __init__(self, _následovník = None, _další = None):
        """Konstruktor vytvoří pomocný prvek seznamu se zadanými daty"""
        self.následovník = _následovník
        self.další = _další

# Seznam pro topologické třídění
class TopoSeznam:
    """Seznam pro topologické třídění"""
    def __init__(self) -> None:
        """Konstruktor vytvoří prázdný seznam se zarážkou"""
        self.hlava = Hlavní()
        self.zarážka = self.hlava
        self.hlava.další = None
        self.kolik = 0

    def je_v_seznamu(self, co): 
        """Vyhledá v seznamu prvek s daným klíčem.
            Když v seznamu není, přidá ho
            a vždy vrátí odkaz na něj"""
        prvek = self.hlava              
        self.zarážka.klíč = co
        while prvek.klíč != co:
            prvek = prvek.další
        if prvek == self.zarážka:        
            self.zarážka = Hlavní()
            self.kolik += 1
            prvek.počet_předchůdců = 0
            prvek.pomocný = None
            prvek.další = self.zarážka
        return prvek                     

    def naplň_seznam_ze_souboru(self, jméno_souboru: str):
        """Přečte obsah vstupního souboru a uloží získané informace 
            do seznamu"""
        with open(jméno_souboru, "r") as soubor:
            řádky = soubor.readlines()
        for řádka in řádky:
            x, y = map(int, řádka.split())
            p = self.je_v_seznamu(x)
            q = self.je_v_seznamu(y)
            t = Pomocný()
            t.následovník = q
            t.další = p.pomocný
            p.pomocný = t
            q.počet_předchůdců += 1

    def najdi_nejmenší(self) ->None:
        """Přeuspořádá seznam"""
        pom = self.hlava
        self.hlava = None
        while pom != self.zarážka:
            pom2 = pom
            pom = pom.další
            if pom2.počet_předchůdců == 0:
                pom2.další = self.hlava
                self.hlava = pom2
                
    def vypiš_seznam(self):
        """Vypíše prvky v pořadí, v němž následující 
           prvek je větší než prvky předcházející 
           nebo s nimi není v žádném vztahu"""
        pom = self.hlava
        while pom is not None:
            print(pom.klíč)
            self.kolik -= 1
            t = pom.pomocný
            pom = pom.další
            while t is not None:
                p = t.následovník
                p.počet_předchůdců -= 1
                if p.počet_předchůdců == 0:
                    p.další = pom
                    pom = p
                t = t.další
        if self.kolik != 0: print("Tato množina není částečně uspořádána")

            
def main()->int:
    s = TopoSeznam()
    s.naplň_seznam_ze_souboru("data.txt")
    s.najdi_nejmenší()
    s.vypiš_seznam()
    return 0

if __name__ == '__main__':
    sys.exit(main())