﻿###################################################
#
#              Obrácený polský zápis 
#
# Převod výrazu do obráceného (reverzního) polského 
# zápisu (RPZ)
# Převáděný výraz je obsahem řetězce vstup a smí 
# obsahovat operandy vyjádřené velkými písmeny 
# anglické abecedy, operátory +, -, *, / 
# a kulaté závorky
###################################################

import sys

global priorita, zásobník, výstup, vstup
priorita = {"+": 1, "-": 1, "*": 2, "/": 2}
zásobník = []
výstup = ""
vstup = "(A+D)*(B-C)+E"

def je_operand(c)->bool:
	"""Test, zda jde o operand;
	   operandem musí být velké písmeno anglické abecedy"""
	return "A" <= c <= "Z"

def je_operátor(c)->bool:
	"""Test, zda jde o operand"""
	return c in priorita
    
# Zpracování operátoru - smí být vložen pouze do
# prázdného zásobníku nebo na "(" nebo na operátor 
# s nižší prioritou, jinak je třeba nejprve operátory 
# v zásobníku, které mají stejnou nebo vyšší prioritu,
# vyjmout a přidat do výstupu
def zpracuj_operátor(c):
	"""Zpracování operátoru"""
	global priorita, zásobník, výstup
	while True:
		if not zásobník:
			zásobník.append(c)
			return
		else:
			pom = zásobník[-1]
			if pom == "(" or priorita[c] > priorita[pom]:
				zásobník.append(c)
				return
			else:
				zásobník.pop()
				výstup += pom

# Při pravé závorce je třeba vyjmout všechny operátory
# až po odpovídající levou závorku a vložit je do výstupu
def zpracuj_pravou_závorku():
	"""Zpracování pravé závorky"""
	global priorita, zásobník, výstup
	while True:
		if not zásobník: 
			raise Exception("Nesouhlas závorek")
		pom = zásobník.pop()
		if pom == '(':
			break
		výstup += pom

def RPZ(vstup)->str:
	"""Převod vstupu do obráceného polského zápisu"""
	global priorita, zásobník, výstup
	for c in vstup:
		if c == ' ': 
			continue
		elif c == '(': 
			zásobník.append(c)
			continue
		elif je_operand(c):
			výstup += c
			continue
		elif je_operátor(c):
			zpracuj_operátor(c)
			continue
		elif c == ')':
			zpracuj_pravou_závorku()
			continue
		raise Exception("Nedovolený znak")
	while zásobník:
		pom = zásobník.pop()
		if pom == "(": raise Exception("Nesouhlas závorek")
		výstup += pom
	return výstup

def main()->int:
	try:
		print(RPZ(vstup))
		return 0
	except Exception as výjimka:
		print("Chyba: " + str(výjimka))
		return 1

if __name__ == '__main__':
	sys.exit(main())
