﻿
###################################################
#
#            Obrácený polský zápis
#
# Převod výrazu do obráceného (reverzního) polského 
# zápisu (RPZ) a jeho vyhodnocení.
# Převáděný výraz je obsahem řetězce "vstup". 
# Operandy jsou vyjádřeny velkými písmeny 
# a jejich hodnoty jsou uloženy ve slovníku
# (dict) "proměnné". Program zde očekává hodnoty typu int.
# Výraz může obsahovat binární operátory +, -, *, / 
# a kulaté závorky.
###################################################

import sys

global priorita, zásobník, výstup, proměnné, vstup

vstup = "(A+D)*(B-C)"
proměnné = {"A": 5, "B": 2, "C": 7, "D": 8}
priorita = {"+": 1, "-": 1, "*": 2, "/": 2}
zásobník = []
výstup = ""

def je_operand(c)->bool:
	"""Test, zda jde o operand;
	   operandem musí být velké písmeno anglické abecedy A ... D"""
	global proměnné
	return c in proměnné  #test, zda slovník obsahuje klíč c

def je_operátor(c)->bool:
	"""Test, zda jde o operátor"""
	return c in priorita
    
#Převod do RPZ:
#Zpracování operátoru - smí být vložen pouze do
#prázdného zásobníku nebo na "(" nebo na operátor 
#s nižší prioritou, jinak je třeba nejprve operátory 
#v zásobníku, které mají stejnou nebo vyšší prioritu,
#vyjmout a přidat do výstupu
def zpracuj_operátor(c):
	"""Zpracování operátoru při převodu výrazu do RPZ"""
	global priorita, zásobník, výstup
	while True:
		if not zásobník:
			zásobník.append(c)
			return
		else:
			pom = zásobník[-1]
			if pom == "(" or priorita[c] > priorita[pom]:
				zásobník.append(c)
				return
			else:
				zásobník.pop()
				výstup += pom

#Při pravé závorce je třeba vyjmout všechny operátory
#až po odpovídající levou závorku a vložit je do výstupu
def zpracuj_pravou_závorku():
	"""Zpracování levé závorky při převodu výrazu do RPZ"""
	global priorita, zásobník, výstup
	while True:
		if not zásobník: 
			raise Exception("Nesouhlas závorek")
		pom = zásobník.pop()
		if pom == '(':
			break
		výstup += pom

def reverzní_polský_zápis(vstup)->str:
	"""Převod zadaného výrazu do obráceného polského zápisu"""
	global priorita, zásobník, výstup
	for c in vstup:
		if c == ' ': 
			continue
		elif c == '(': 
			zásobník.append(c)
			continue
		elif je_operand(c):
			výstup += c
			continue
		elif je_operátor(c):
			zpracuj_operátor(c)
			continue
		elif c == ')':
			zpracuj_pravou_závorku()
			continue
		raise Exception("Nedovolený znak")
	while zásobník:          #dokud je zásobník neprázdný
		pom = zásobník.pop()
		if pom == "(": raise Exception("Nesouhlas závorek")
		výstup += pom;
	return výstup

def vypočti_výraz(operand1, operand2, operátor): 
	"""Vyhodnocení dílčího výrazu zadaného operátorem a dvěma operandy"""
	match operátor:
		case "+": return operand2 + operand1
		case "-": return operand2 - operand1
		case "*": return operand2 * operand1
		case "/": return operand2 / operand1

def výpočet(rpz: str)->int:
	"""Vyhodnocení výrazu zapsaného v RPZ"""
	global zásobník
	zásobník = []                   #Nejprve vyprázdníme zásobník
	for c in rpz:
		if je_operand(c):           #Operandy vkládáme do zásobníku
			zásobník.append(proměnné[c])
			continue
		else:                       #Operátor aplikujeme na hodnoty ne vrcholu zásobníku
			zásobník.append(vypočti_výraz(zásobník.pop(), zásobník.pop(), c))
	výsledek = zásobník.pop()       #a výsledek uložíme na vrchol zásobníku
	if zásobník: raise Exception("Neprázdný zásobník po skončení výpočtu")
	return výsledek


def main()->int:
	try:
		print(f"Zadaný výraz: {vstup}")
		rpz = reverzní_polský_zápis(vstup)
		print(f"Převedeno do RPZ: {rpz}")
		print()
		print(f"Hodnota je {výpočet(rpz)}")
		print(f"Hodnoty proměnných: {proměnné}")
		return 0
	except Exception as výjimka:
		print("Chyba: " + str(výjimka))
		return 1

if __name__ == '__main__':
    sys.exit(main())
