﻿################################################################
# 
#       Výpočet sinu pomocí Talorova rozvoje v 0
#
# Program vypočte a vypíše kosinus 0, π/2, π a 100 počítaný
# jako součet Taylorovy řady pro kosinus v nule. Sčítání skončí,
# když bude sčítanec menší než dané epsilon (zde 1E-5). První 
# tři hodnoty budou blízké očekávaným (s chybou nepřesahující 
# epsilon), čtvrtá je díky zaokrouhlovacím chybám zcela nesmyslná. 
#
# Vedle toho vypíše vypočte a vypíše kosinus týchž hodnot pomocí 
# podobné funkce, která však využívá typ decimal s přesností 
# 50 platných cifer. Pro kontrolu vypíše i kosimus hodnoty 100 
# vypočtený pomocí knihovní funkce math.cos. Výsledek se tentokrát 
# ve všech případech liší od správné hodnoty o méně než epsilon.
#
# Program dále vypočte a vypíše sin i * π pro i = 1, ..., 20 
# počítaný opět # jako součet Taylorovy řady pro sinus v nule 
# pomocí typu float. Výsledek by měl být vždy blízký nule;
# zaokrouhlovací chyby však způsobí, že pro větší i brzy 
# ztratí smysl.
#
# (Skutečná hodnota sinu i*π je 0, avšak pi není v knihovně
# přesně a knihovní funkce sin() také nepočítá absolutně přesně, 
# takže vypsané hodnoty knihovní funkce budou nenulové, 
# ale velmi malé.)
###############################################################

import decimal
from math import sin, cos, pi
import sys

from decimal import *

def sinus(x:float, eps: float)->float:      
    """Implementace funkce sinus založená na sčítání Taylorovy řady v bodě 0"""
    k, výsledek, sčítanec = 0, 0.0, x       #k je sčítací index
    while abs(sčítanec) > eps:
        výsledek = výsledek + sčítanec
        k += 1
        sčítanec = -sčítanec*x*x/(2*k + 1)/(2*k)
    return výsledek

def kosinus(x: float, eps: float)->float:
    """Implementace funkce kosinus založená na sítání Taylorovy řady v bodě 0"""
    k, výsledek, sčítanec = 0, 1.0, -x*x/2.0       #k je sčítací index
    while abs(sčítanec) > eps:
        výsledek += sčítanec
        k += 1
        sčítanec = -sčítanec*x*x/(2*k+1)/(2*k + 2)
    return výsledek


def kosinus_decimal(x: float, eps: float)->Decimal:
    """Implementace funkce kosinus založená na sítání Taylorovy řady 
       v bodě 0 s použitím typu Decimal"""
    k = 0
    výsledek: Decimal = Decimal(1)
    x_dec: Decimal = Decimal(x)
    sčítanec: Decimal = -x_dec*x_dec/Decimal(2*k+1)/Decimal(2*k+2)
    while abs(sčítanec) > eps:
        výsledek += sčítanec
        k += 1
        sčítanec = -sčítanec*x_dec*x_dec/Decimal(2*k+1)/Decimal(2*k+2)
    return výsledek

def main()->int:
    eps = 1e-5 
    π = pi
    
    print(f"kosinus(0, eps) = {kosinus(0, eps)}")
    print(f"kosinus(π/2, eps) = {kosinus(π/2, eps)}")
    print(f"kosinus(π, eps) = {kosinus(π, eps)}")
    print(f"kosinus(100, eps) = {kosinus(100, eps)}")

    getcontext().prec = 50
    print(f"kosinus_decimal(0, eps) = {kosinus_decimal(0, eps)}")
    print(f"kosinus_decimal(π/2, eps) = {kosinus_decimal(π/2, eps)}")
    print(f"kosinus_decimal(π, eps) = {kosinus_decimal(π, eps)}")
    print(f"kosinus_decimal(100, eps) = {kosinus_decimal(100, eps)}")
    print(f"cos(100) = {cos(100)}")

    for i in range(1, 21):
        print("Vypočtený sinus %2d * π je %10.3E, ve skutečnosti je %10.3E" % (i, sinus(i*π, eps), sin(i*π)))
    return 0

if __name__ == '__main__':
    sys.exit(main())

